
#define	 F_CPU	16000000UL
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include "ili9341.h"
#include "ili9341gfx.h"
#include <math.h>
#define	 POINTCOLOUR	PINK

#define NORTH	0
#define EAST 	1
#define SOUTH	2
#define WEST	3

#define OBSTACLE 			0
#define FOOD 				1
#define CURE				2
#define GAME_END 			1
#define FOOD_POINTS 		10
#define LEVEL_X 			60
#define OPTION_SELECTOR_X 	100
#define EASY_Y 				8*15
#define MEDIUM_Y 			8*20
#define DIFFICULT_Y 		8*25
#define SNAKE_MAX_DIM		50
#define WALL_DIM			5
#define EASY_MAX_OBSTACLES	5
#define MED_MAX_OBSTACLES	15
#define HARD_MAX_OBSTACLES	30
#define INIT_SIZE			8

extern uint16_t vsetx,vsety,vactualx,vactualy,isetx,isety,iactualx,iactualy;
static FILE mydata = FDEV_SETUP_STREAM(ili9341_putchar_printf, NULL, _FDEV_SETUP_WRITE);

/* snake coordinates for each rectangle composing the snake */
uint16_t pos_x[SNAKE_MAX_DIM];
uint16_t pos_y[SNAKE_MAX_DIM];
/*coordinates for obstacles */
uint16_t obs_x[HARD_MAX_OBSTACLES];
uint16_t obs_y[HARD_MAX_OBSTACLES];
/* nr of displayed obstacles */
uint8_t nr_obs;
/* intital coordinates for the head ractangle */
uint16_t head_x = ILI9341_TFTHEIGHT / 2 ;
uint16_t head_y = ILI9341_TFTWIDTH / 2;
/* a rectangle is 8*8 pixels */
uint8_t square_size = 8;
/* distance between 2 consecutive rectangles */
uint8_t distance = 2;
/* intial size of the snake */
int size = 8;
uint8_t finished;
uint16_t score;
uint8_t increased = 0;

/* a flag that establishes when level was selected */
uint8_t level_sel = 0;
int16_t choice = 0;

uint8_t direction = SOUTH;
uint16_t food_x;
uint16_t food_y;
uint16_t cure_x, cure_y;

uint8_t food_present;
uint8_t cure_present;
uint8_t max_obs;

int counter;
int cure_counter;

// function to draw all other 7 pixels
// present at symmetric position
void drawCircle(uint16_t xc, uint16_t yc, uint16_t x, uint16_t y, uint16_t color)
{
	ili9341_drawpixel(xc+x, yc+y, color);
    ili9341_drawpixel(xc-x, yc+y, color);
    ili9341_drawpixel(xc+x, yc-y, color);
    ili9341_drawpixel(xc-x, yc-y, color);
    ili9341_drawpixel(xc+y, yc+x, color);
    ili9341_drawpixel(xc-y, yc+x, color);
    ili9341_drawpixel(xc+y, yc-x, color);
    ili9341_drawpixel(xc-y, yc-x, color);
}

// Function for circle-generation
// using Bresenham's algorithm
void circleBres(uint16_t xc, uint16_t yc, uint16_t r, uint16_t color)
{
    uint16_t x = 0, y = r;
    int d = 3 - 2 * r;
    while (y >= x)
    {
        drawCircle(xc, yc, x, y, color);
        x++;

        if (d > 0)
        {
            y--; 
            d = d + 4 * (x - y) + 10;
        }
        else
            d = d + 4 * x + 6;
        drawCircle(xc, yc, x, y, color);
        _delay_ms(5);
    }
}

void draw_food() {
	if (food_present == 0) {
    	circleBres(food_x, food_y, 8, YELLOW);
		circleBres(food_x, food_y, 5, YELLOW);
		food_present = 1;
	}
}

void draw_cure() {
	if (cure_present == 1) {
		ili9341_fillrect(cure_x, cure_y, square_size, square_size, BLUE);
		_delay_ms(2);
	}
}

void get_coordinates(uint8_t type) {
	uint16_t x, y;
	uint8_t ok = 0;
	uint8_t i;

	if (type == FOOD && food_present == 1)
		return;
	while (ok != 1) {
		ok = 1;
		x = rand() % (ILI9341_TFTWIDTH - 80)  + WALL_DIM + distance;
		y = rand() % (ILI9341_TFTHEIGHT - 80) + WALL_DIM + distance;

		/* check if spot is occupied by snake*/
		for (i = 0; i < size; i++) {
			if ((pos_x[i] - x) <= distance || (x - pos_x[i]) <= distance)
				if ((pos_y[i] - y) <= distance || (y - pos_y[i]) <= distance)
					ok = 0;
		}

		/* check if spot is occupied by obstacles */
		for (i = 0; i < nr_obs; i ++) {
			if ((obs_x[i] - x) <= distance || (x - obs_x[i]) <= distance)
				if ((obs_y[i] - y) <= distance || (y - obs_y[i]) <= distance)
					ok = 0;
		}
	}

	if (food_present == 0 && type == FOOD) {
		food_x = x;
		food_y = y;
		return;
	}

	if (type == CURE && cure_present == 0) {
		cure_x = x;
		cure_y = y;
		cure_present = 1;
		return;
	}

	if (type == OBSTACLE){
		nr_obs++;
		obs_x[nr_obs - 1] = x;
		obs_y[nr_obs - 1] = y;
	}

	return;
}

void draw_obstacles() {
	int i;

	for (i = 0; i < nr_obs; i++)
		ili9341_fillrect(obs_x[i], obs_y[i], square_size, square_size, RED);
	_delay_ms(3);
}

void draw_vertical_walls() {
	/* left vertical wall */
	ili9341_fillrect(0, 0,WALL_DIM, ILI9341_TFTHEIGHT, RED);
	_delay_ms(2);
	/* right vertical wall */
	ili9341_fillrect(ILI9341_TFTWIDTH - WALL_DIM, 0,
		WALL_DIM, ILI9341_TFTHEIGHT, RED);
	_delay_ms(2);
}

void draw_horizontal_walls() {
	/* upper horizontal wall */
	ili9341_fillrect(0,0,ILI9341_TFTWIDTH, WALL_DIM, RED);
	/* lower horizontal wall */
	ili9341_fillrect(0, ILI9341_TFTHEIGHT - WALL_DIM,
		ILI9341_TFTWIDTH, WALL_DIM, RED);
}

void draw_walls() {
	switch(choice) {
		case 0:
			return;
		case 1:
			draw_vertical_walls();
			return;
		case 2:
			draw_vertical_walls();
			draw_horizontal_walls();
			return;
	}
}

int get_y_for_level(int level)
{
	switch(level){
		case 0:
			return EASY_Y;
		case 1:
			return MEDIUM_Y;
		case 2:
			return DIFFICULT_Y;
		default:
			return EASY_Y;
	}

	return EASY_Y;
}

ISR(PCINT0_vect)
{
	if (finished == 0) {
	    if ((PINA & (1 << PA0)) == 0)
	    {
	    	if (level_sel == 0) {
		        ili9341_fillrect(2.5 * LEVEL_X, get_y_for_level(choice), 10, 10, OLIVE);
		        _delay_ms(2);
				choice = (choice + 1) % 3;
				ili9341_fillrect(2.5 * LEVEL_X, get_y_for_level(choice), 10, 10, WHITE);
				_delay_ms(2);
			} else if (direction != NORTH)
				direction = SOUTH;
	    }

	    if ((PINA & (1 << PA1)) == 0)
	    {
	    	if (direction != EAST)
	    		direction = WEST;
	    }

	    if ((PINA & (1 << PA2)) == 0) 
	    {
	    	if (level_sel == 0) {
		    	ili9341_fillrect(2.5 * LEVEL_X, get_y_for_level(choice), 10, 10, OLIVE);
		    	_delay_ms(2);
		    	choice -= 1;
		    	if (choice < 0)
		    		choice = 2;
		    	ili9341_fillrect(2.5 * LEVEL_X, get_y_for_level(choice), 10, 10, WHITE);
				_delay_ms(2);
			} else if (direction != SOUTH)
				direction = NORTH;
	    }

	    if ((PINA & (1 << PA3)) == 0 && direction != WEST)
	    {
	    	direction = EAST;
	    }

		if ((PINA & (1 << PA4)) == 0 && level_sel == 0) {
			level_sel = 1;
			/* clear screen so that the game can start */
			ili9341_clear(OLIVE);
		}  

		_delay_ms(5);
	}

	if (level_sel == 0 || finished == 1)
    	_delay_ms(100);
}

void move_snake() {
	int deltaX = 0;
	int deltaY = 0;
	uint16_t diffx, diffy;
	int i;

	/* compute distance for x and y coordinates regarding to the movement
	 * direction
	 */
	switch(direction) {
		case NORTH:
			deltaY = - square_size - distance;
			break;
		case SOUTH:
			deltaY = square_size + distance;
			break;
		case EAST:
			deltaX = square_size + distance;
			break;
		case WEST:
			deltaX = - square_size - distance;
			break;
	}

	if (score % 30 == 0 && score > 0 &&  increased == 0) {
		size += 1;
		increased = 1;
	} else {
		/* erase last coordinate */
		ili9341_fillrect(pos_x[size - 1], pos_y[size - 1], square_size, square_size, OLIVE);
		_delay_ms(2);
	}


	/* iterate through coordinates to change position */
	for (i = size - 1; i >= 1; i--) {
		pos_x[i] = pos_x[i - 1];
		pos_y[i] = pos_y[i - 1];
	}

	/* update x coordinate for head */
	pos_x[0] = (head_x + deltaX);
	/* boundry checks */
	if(pos_x[0] <= 0 && choice == 0)
		pos_x[0] = ILI9341_TFTWIDTH - square_size - distance;
	else if (pos_x[0] > (ILI9341_TFTWIDTH - square_size - distance) && choice == 0)
		pos_x[0] = 0;
	/* update y coordinate for head */
	pos_y[0] = head_y + deltaY;
	if (pos_y[0] < square_size + distance && choice < 2)
		pos_y[0] = ILI9341_TFTHEIGHT -square_size - distance;
	else if (pos_y[0] > ILI9341_TFTHEIGHT -square_size - distance &&
		choice < 2)
		pos_y[0] = 0;

	/* final coordinates */
	head_x = pos_x[0];
	head_y = pos_y[0];

	if (head_x > food_x)
		diffx = head_x - food_x;
	else
		diffx = food_x - head_x;

	if (head_y > food_y)
		diffy = head_y - food_y;
	else
		diffy = food_y - head_y;

	if ((diffx * diffx + diffy * diffy <= 7 * 7) && food_present == 1) {
		/* eat food */
		circleBres(food_x,food_y, 8, OLIVE);
		circleBres(food_x, food_y, 5, OLIVE);
		food_x = head_x;
		food_y = head_y;
		score += FOOD_POINTS;
		food_present = 0;
		if (score % 30 != 0 && score > 0)
			increased = 0;
	}

	/* check if cure was taken */
	if (cure_present == 1 && head_x >= cure_x && 
		head_x <= (cure_x + square_size) && head_y >= cure_y &&
		head_y <= (cure_y + square_size)) {
		/*erase cure from map */
		ili9341_fillrect(cure_x, cure_y, square_size, square_size, OLIVE);
		_delay_ms(2);
		/* erase tail once more */
		ili9341_fillrect(pos_x[size - 1], pos_y[size - 1], square_size, square_size, OLIVE);
		/* decrease size of the snake */
		size -= 1;
		cure_present = 0;
	}

	return;
}

uint8_t horizontal_collision() {
	if (choice != 2)
		return 0;
	if (head_y <= (WALL_DIM + distance))
		return GAME_END;
	if (head_y >= ILI9341_TFTHEIGHT - WALL_DIM)
		return GAME_END;
	return 0;
}

uint8_t vertical_collision() {
	if (choice < 1)
		return 0;
	if (head_x <= WALL_DIM)
		return GAME_END;
	if (head_x >= ILI9341_TFTWIDTH - WALL_DIM)
		return GAME_END;
	return 0;
}

uint8_t is_game_end() {
	int i;

	if (vertical_collision() == GAME_END || 
		horizontal_collision() == GAME_END)
		return GAME_END;

	/* check collisions*/
	for (i = 0; i < size; i++) {
		if (((head_x - pos_x[i]) < distance || (pos_x[i] - head_x) < distance) &&
			 ((head_y - pos_y[i]) < distance || (pos_y[i] - head_y) < distance) && i > 0)
			return GAME_END;

		if (i < nr_obs) {
			if (head_x >= obs_x[i] && head_x <= (obs_x[i] + square_size) && head_y >= obs_y[i] &&
				head_y <= (obs_y[i] + square_size))
				return GAME_END;
		}
	}

	return 0;
}

/* draw one rectangle from the snake */
void draw_snake_piece(uint16_t x, uint16_t y, uint16_t color)
{
	ili9341_fillrect(x, y, square_size, square_size, BLACK);
	//_delay_ms(2);
}

/* draw the entire snake */
void draw_snake() {
	int i = 0;
	move_snake();
	for (i = 0; i < size; i++) {
		draw_snake_piece(pos_x[i], pos_y[i], BLACK);
	}
	_delay_ms(5);
}


void print_text(uint16_t cursor_x, uint16_t cursor_y, uint8_t text_size,
	uint16_t text_color, char *text) {
	ili9341_setcursor(cursor_x, cursor_y);
	_delay_ms(2);
	ili9341_settextcolour(text_color, OLIVE);
	_delay_ms(2);
	ili9341_settextsize(text_size);
	_delay_ms(2);
	printf(text);
}

void print_menu() {
	uint8_t choice = 0;
	ili9341_init(); //initial driver setup to drive ili9341
	ili9341_clear(OLIVE); /* set initial background color */
	ili9341_setRotation(1);//rotate screen
	_delay_ms(2);

	print_text(0, 0, 5, WHITE, "Snake++");
	print_text(0, 40, 2, WHITE, "Meirosu Cristina 332CA");
	print_text(0, 80, 2, WHITE, "Selectati nivelul: ");
	print_text(LEVEL_X, EASY_Y, 2, WHITE, "Usor");
	print_text(LEVEL_X, MEDIUM_Y, 2, WHITE, "Mediu");
	print_text(LEVEL_X, DIFFICULT_Y, 2, WHITE, "Dificil");


	/* mark "easy" as default option */
	ili9341_fillrect(2.5 * LEVEL_X, get_y_for_level(choice), 10, 10, WHITE);
	return;
}

void init_parameters() {
	int i;

	if (choice == 0) {
		max_obs = EASY_MAX_OBSTACLES;
	} else if (choice == 1) {
		max_obs = MED_MAX_OBSTACLES;
	} else {
		max_obs = HARD_MAX_OBSTACLES;
	}

	counter = 0;
	cure_counter = 0;
	cure_present = 0;
	food_present = 0;
	score = 0;
	size = INIT_SIZE;
	nr_obs = 0;
	head_x = ILI9341_TFTHEIGHT / 2 ;
	head_y = ILI9341_TFTWIDTH / 2;

	 /* init snake_coordinates */
    pos_x[0] = head_x;
    pos_y[0] = head_y;
    for (i = 1; i < size; i++) {
    	pos_x[i] = pos_x[i - 1];
    	pos_y[i] = pos_y[i - 1] - square_size - distance;
    } 
}

int main(void)
{
	int i = 0;
	uint8_t menu_printed = 0;
	uint8_t init_done = 0;
	char buff[10];
	srand(1);

	//uint8_t timer_set = 0;
	stdout = & mydata;

	/* Setez pinii PA0, PA1, PA2, PA3, PA4 ca butoane de intrare */
	DDRA &= ~(1 << PA0) & ~(1 << PA1) & ~(1 << PA2) & ~(1 << PA3) & ~(1 << PA4);

	/* Activez rezistenta de pull-up pentru pinii PA0, PA1, PA2, PA3, PA4 */
	PORTA |= (1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4);

	/* allow interrupts */
	sei();

	/* Activez PIN CHANGE INTERRUPT ENABLE pentru PCINT7:0 (portul A) */
    PCICR |= (1 << PCIE0);
    /* Activez intreruperea pentru butoanele PA0, PA2, PA4 */
    PCMSK0 |= (1 << PCINT0) | (1 << PCINT2) | (1 << PCINT4) | (1 << PCINT1) |  (1 << PCINT3);

	while(1) {
		if (finished == 1) {
			continue;
		}
		if (menu_printed == 0) {
			print_menu();
			menu_printed = 1;
		}
		if (level_sel == 1) {
			if (init_done == 0) {
				init_parameters();
				init_done = 1;
			}
			finished = is_game_end();
			while (finished != GAME_END) {
				get_coordinates(FOOD);
				draw_food();
				draw_walls();
				if (nr_obs > 0) {
					draw_obstacles();
				}
				draw_snake();
				draw_cure();
				/* print score */
				memset(buff, 0, 10);
				sprintf(buff, "Score: %d", score);
				print_text(WALL_DIM + 1, WALL_DIM + 1, 1, WHITE, &buff);
				if (counter % 30 == 0 && nr_obs < max_obs) {
					get_coordinates(OBSTACLE);
				}
				if (cure_counter % 80 == 0 && cure_counter > 0) {
					get_coordinates(CURE);
				}
				counter++;
				cure_counter += 1;
				finished = is_game_end();
				_delay_ms(2);
			}
			ili9341_clear(OLIVE);
			_delay_ms(2);
			memset(buff, 0, 10);
			sprintf(buff, "Final score: %d", score);
			print_text(20, 120, 3, PURPLE, &buff);
			print_text(LEVEL_X, DIFFICULT_Y, 2, PURPLE, 
				"Apasati OK");
			init_done = 0;
			menu_printed = 0;
			level_sel = 0;
			while (1) {
				if (((PINA & (1 << PA4)) == 0)) {
					finished = 0;
					break;
				}
			}	
		} else {
				_delay_ms(10);
		}
	}

	return 0;
}